INT = 'int'
FLOAT = 'float'
BOOL = 'bool'
STR = 'str'
DICT = 'dict'
DICT_OF_DICTS = 'dict_of_dicts'
URI = 'uri'
AUTH = 'auth'
LIST = 'list'
GCS_PATH = 'gcs_path'
S3_PATH = 's3_path'
AZURE_PATH = 'azure_path'
PATH = 'path'
METRIC = 'metric'
METADATA = 'metadata'

VALUES = {
    INT,
    FLOAT,
    BOOL,
    STR,
    DICT,
    DICT_OF_DICTS,
    URI,
    AUTH,
    LIST,
    GCS_PATH,
    S3_PATH,
    AZURE_PATH,
    PATH,
    METRIC,
    METADATA
}



from unittest import TestCase

from rhea import rhea_parser
from rhea.constants import NO_VALUE_FOUND
from rhea.exceptions import RheaError
from rhea.specs import AuthSpec, GCSSpec, S3Spec, UriSpec, WasbsSpec


class TestRheaParser(TestCase):

    def test_get_boolean(self):
        value = rhea_parser.get_boolean(key='bool_key_1', value="1")
        self.assertEqual(value, True)

        value = rhea_parser.get_boolean(key='bool_key_2', value="true")
        self.assertEqual(value, True)

        value = rhea_parser.get_boolean(key='bool_key_2', value=True)
        self.assertEqual(value, True)

        value = rhea_parser.get_boolean(key='bool_key_3', value="0")
        self.assertEqual(value, False)

        value = rhea_parser.get_boolean(key='bool_key_4', value="false")
        self.assertEqual(value, False)

        value = rhea_parser.get_boolean(key='bool_key_4', value=False)
        self.assertEqual(value, False)

        value = rhea_parser.get_boolean(
            key='bool_list_key_1',
            value=[False, "false", True, "true", "1", "0"],
            is_list=True)
        self.assertEqual(value, [False, False, True, True, True, False])

        with self.assertRaises(RheaError):
            rhea_parser.get_boolean(key='bool_error_key_1', value="null")

        with self.assertRaises(RheaError):
            rhea_parser.get_boolean(key='bool_error_key_1', value=None)

        with self.assertRaises(RheaError):
            rhea_parser.get_boolean(key='bool_error_key_2', value="foo")

        with self.assertRaises(RheaError):
            rhea_parser.get_boolean(key='bool_error_key_3', value=0)

        with self.assertRaises(RheaError):
            rhea_parser.get_boolean(key='bool_error_key_4', value=1)

        with self.assertRaises(RheaError):
            rhea_parser.get_boolean(key='bool_error_key_5', value="")

        with self.assertRaises(RheaError):
            rhea_parser.get_boolean(key='bool_list_key_1',
                                    value=[False, "false", True, "true", "1", "0"])

        with self.assertRaises(RheaError):
            rhea_parser.get_boolean(key='bool_list_error_key_2',
                                    value=[False, 1, 0],
                                    is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_boolean(key='bool_list_error_key_1',
                                    value=[False, "false", True, "true", "1", "0", "foo"],
                                    is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_boolean(key='bool_key_1',
                                    value="1",
                                    is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_boolean(key='bool_key_2',
                                    value=True,
                                    is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_boolean(key='bool_non_existing_key',
                                    value=None)

        with self.assertRaises(RheaError):
            rhea_parser.get_boolean(key='bool_non_existing_key',
                                    value=NO_VALUE_FOUND)

        with self.assertRaises(RheaError):
            rhea_parser.get_boolean(key='bool_non_existing_key',
                                    value=None,
                                    is_list=True)

        self.assertEqual(rhea_parser.get_boolean(
            key='bool_non_existing_key',
            value=None,
            is_optional=True), None)
        self.assertEqual(rhea_parser.get_boolean(
            key='bool_non_existing_key',
            value=None,
            is_optional=True,
            default=True), True)

        self.assertEqual(rhea_parser.get_boolean(
            key='bool_non_existing_key',
            value=None,
            is_list=True,
            is_optional=True), None)
        self.assertEqual(rhea_parser.get_boolean(
            key='bool_non_existing_key',
            value=None,
            is_list=True,
            is_optional=True,
            default=[True, False]),
            [True, False])

    def test_get_int(self):
        value = rhea_parser.get_int(key='int_key_1', value=123)
        self.assertEqual(value, 123)

        value = rhea_parser.get_int(key='int_key_2', value="123")
        self.assertEqual(value, 123)

        value = rhea_parser.get_int(key='int_list_key_1',
                                    value=["123", 124, 125, "125"],
                                    is_list=True)
        self.assertEqual(value, [123, 124, 125, 125])

        value = rhea_parser.get_int(key='int_list_key_1',
                                    value='["123", 124, 125, "125"]',
                                    is_list=True)
        self.assertEqual(value, [123, 124, 125, 125])

        with self.assertRaises(RheaError):
            rhea_parser.get_int(key='int_error_key_1', value='null')

        with self.assertRaises(RheaError):
            rhea_parser.get_int(key='int_error_key_1', value=None)

        with self.assertRaises(RheaError):
            rhea_parser.get_int(key='int_error_key_2', value='')

        with self.assertRaises(RheaError):
            rhea_parser.get_int(key='int_error_key_3', value='foo')

        with self.assertRaises(RheaError):
            rhea_parser.get_int(key='int_list_key_1', value=["123", 124, 125, "125"])

        with self.assertRaises(RheaError):
            rhea_parser.get_int(key='int_list_error_key_1',
                                value=["123", 124, 125, "125", None],
                                is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_int(key='int_list_error_key_2',
                                value=["123", 1.24, 125, "125"],
                                is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_int(key='int_list_error_key_3',
                                value=["123", 1.24, 125, "foo"],
                                is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_int(key='int_key_1', value=125, is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_int(key='int_key_2', value="125", is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_int(key='int_non_existing_key', value=None)

        with self.assertRaises(RheaError):
            rhea_parser.get_int(key='int_non_existing_key', value=NO_VALUE_FOUND)

        self.assertEqual(rhea_parser.get_int(key='int_non_existing_key',
                                             value=None,
                                             is_optional=True), None)
        self.assertEqual(rhea_parser.get_int(key='int_non_existing_key',
                                             value=None,
                                             is_optional=True,
                                             default=34), 34)

        self.assertEqual(rhea_parser.get_int(key='int_non_existing_key',
                                             value=None,
                                             is_list=True,
                                             is_optional=True), None)
        self.assertEqual(rhea_parser.get_int(key='int_non_existing_key',
                                             value=None,
                                             is_list=True,
                                             is_optional=True,
                                             default=[34, 1]), [34, 1])

    def test_get_float(self):
        value = rhea_parser.get_float(key='float_key_1', value=1.23)
        self.assertEqual(value, 1.23)

        value = rhea_parser.get_float(key='float_key_2', value="1.23")
        self.assertEqual(value, 1.23)

        value = rhea_parser.get_float(key='float_key_3', value="123")
        self.assertEqual(value, 123)

        value = rhea_parser.get_float(key='float_list_key_1',
                                      value=[1.23, 13.3, "4.4", "555", 66.0],
                                      is_list=True)
        self.assertEqual(value, [1.23, 13.3, 4.4, 555., 66.])

        value = rhea_parser.get_float(key='float_list_key_1',
                                      value='[1.23, 13.3, "4.4", "555", 66.0]',
                                      is_list=True)
        self.assertEqual(value, [1.23, 13.3, 4.4, 555., 66.])

        with self.assertRaises(RheaError):
            rhea_parser.get_float(key='float_error_key_1', value=None)

        with self.assertRaises(RheaError):
            rhea_parser.get_float(key='float_error_key_1', value='null')

        with self.assertRaises(RheaError):
            rhea_parser.get_float(key='float_error_key_2', value="")

        with self.assertRaises(RheaError):
            rhea_parser.get_float(key='float_error_key_3', value="foo")

        with self.assertRaises(RheaError):
            rhea_parser.get_float(key='float_error_key_4', value=123)

        with self.assertRaises(RheaError):
            rhea_parser.get_float(key='float_list_key_1', value=[1.23, 13.3, "4.4", "555", 66.0])

        with self.assertRaises(RheaError):
            rhea_parser.get_float(key='float_list_error_key_1', value=None, is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_float(key='float_list_error_key_2', value="", is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_float(key='float_list_error_key_3', value="foo", is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_float(key='float_key_1', value=213, is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_float(key='float_key_2', value=[1.23, 13.3, 66], is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_float(key='float_non_existing_key', value=NO_VALUE_FOUND)

        with self.assertRaises(RheaError):
            rhea_parser.get_float(key='float_non_existing_key', value=[1.23, 13.3, "foo"])

        with self.assertRaises(RheaError):
            rhea_parser.get_float(key='float_non_existing_key',
                                  value=[1.23, 13.3, None],
                                  is_list=True)

        self.assertEqual(rhea_parser.get_float(key='float_non_existing_key',
                                               value=None,
                                               is_optional=True), None)
        self.assertEqual(rhea_parser.get_float(key='float_non_existing_key',
                                               value=None,
                                               is_optional=True,
                                               default=3.4), 3.4)

        self.assertEqual(rhea_parser.get_float(key='float_non_existing_key',
                                               value="null",
                                               is_list=True,
                                               is_optional=True), None)
        self.assertEqual(rhea_parser.get_float(key='float_non_existing_key',
                                               value=None,
                                               is_list=True,
                                               is_optional=True,
                                               default=[3.4, 1.2]), [3.4, 1.2])

    def test_get_string(self):
        value = rhea_parser.get_string(key='string_key_1', value="123")
        self.assertEqual(value, "123")

        value = rhea_parser.get_string(key='string_key_2', value="1.23")
        self.assertEqual(value, "1.23")

        value = rhea_parser.get_string(key='string_key_3', value="foo")
        self.assertEqual(value, "foo")

        value = rhea_parser.get_string(key='string_key_4', value="")
        self.assertEqual(value, "")

        value = rhea_parser.get_string(key='string_list_key_1',
                                       value=["123", "1.23", "foo", ""],
                                       is_list=True)
        self.assertEqual(value, ["123", "1.23", "foo", ""])

        value = rhea_parser.get_string(key='string_list_key_1',
                                       value='["123", "1.23", "foo", ""]',
                                       is_list=True)
        self.assertEqual(value, ["123", "1.23", "foo", ""])

        with self.assertRaises(RheaError):
            rhea_parser.get_string(key='string_error_key_1', value=None)

        with self.assertRaises(RheaError):
            rhea_parser.get_string(key='string_error_key_2', value=123)

        with self.assertRaises(RheaError):
            rhea_parser.get_string(key='string_error_key_3', value=1.23)

        with self.assertRaises(RheaError):
            rhea_parser.get_string(key='string_error_key_4', value=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_string(key='string_list_key_1', value=["123", "1.23", "foo", ""])

        with self.assertRaises(RheaError):
            rhea_parser.get_string(key='string_list_error_key_1',
                                   value=["123", 123],
                                   is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_string(key='string_list_error_key_2',
                                   value=["123", 12.3],
                                   is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_string(key='string_list_error_key_3',
                                   value=["123", None],
                                   is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_string(key='string_list_error_key_4',
                                   value=["123", False],
                                   is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_string(key='string_key_3', value="foo", is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_string(key='string_key_4', value="", is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_string(key='string_non_existing_key', value=None)

        with self.assertRaises(RheaError):
            rhea_parser.get_string(key='string_non_existing_key', value=NO_VALUE_FOUND)

        with self.assertRaises(RheaError):
            rhea_parser.get_string(key='string_non_existing_key', value=None, is_list=True)

        self.assertEqual(rhea_parser.get_string(key='string_non_existing_key',
                                                value=None,
                                                is_optional=True), None)
        self.assertEqual(rhea_parser.get_string(key='string_non_existing_key',
                                                value=None,
                                                is_optional=True,
                                                default='foo'), 'foo')

        self.assertEqual(rhea_parser.get_string(key='string_non_existing_key',
                                                value=None,
                                                is_list=True,
                                                is_optional=True), None)
        self.assertEqual(rhea_parser.get_string(key='string_non_existing_key',
                                                value=None,
                                                is_list=True,
                                                is_optional=True,
                                                default=['foo', 'bar']), ['foo', 'bar'])

    def test_get_dict(self):
        value = rhea_parser.get_dict(key='dict_key_1',
                                     value={"key1": "foo", "key2": 2, "key3": False, "key4": "1"})
        self.assertEqual(value, {"key1": "foo", "key2": 2, "key3": False, "key4": "1"})

        value = rhea_parser.get_dict(key='dict_key_1',
                                     value='{"key1": "foo", "key2": 2, "key3": false, "key4": "1"}')
        self.assertEqual(value, {"key1": "foo", "key2": 2, "key3": False, "key4": "1"})

        value = rhea_parser.get_dict(
            key='dict_key_1',
            value="{\"key1\": \"foo\", \"key2\": 2, \"key3\": false, \"key4\": \"1\"}")
        self.assertEqual(value, {"key1": "foo", "key2": 2, "key3": False, "key4": "1"})

        value = rhea_parser.get_dict(key='dict_list_key_1',
                                     value=[
                                         {"key1": "foo", "key2": 2, "key3": False, "key4": "1"},
                                         {"key3": True, "key4": "2"},
                                         {"key1": False, "key2": "3"}],
                                     is_list=True)
        self.assertEqual(value, [
            {"key1": "foo", "key2": 2, "key3": False, "key4": "1"},
            {"key3": True, "key4": "2"},
            {"key1": False, "key2": "3"}
        ])

        with self.assertRaises(RheaError):
            rhea_parser.get_dict(key='dict_error_key_1', value="foo")

        with self.assertRaises(RheaError):
            rhea_parser.get_dict(key='dict_error_key_2', value=1)

        with self.assertRaises(RheaError):
            rhea_parser.get_dict(key='dict_error_key_3', value=False)

        with self.assertRaises(RheaError):
            rhea_parser.get_dict(key='dict_error_key_4', value=["1", "foo"])

        with self.assertRaises(RheaError):
            rhea_parser.get_dict(key='dict_list_key_1', value=["123", {"key3": True}])

        with self.assertRaises(RheaError):
            rhea_parser.get_dict(key='dict_list_error_key_1',
                                 value=["123", {"key3": True}],
                                 is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_dict(key='dict_list_error_key_2',
                                 value=[{"key3": True}, 12.3],
                                 is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_dict(key='dict_list_error_key_3',
                                 value=[{"key3": True}, None],
                                 is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_dict(key='dict_list_error_key_4',
                                 value=[{"key3": True}, "123", False],
                                 is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_dict(key='dict_key_1',
                                 value={"key1": "foo", "key2": 2, "key3": False, "key4": "1"},
                                 is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_dict(key='dict_non_existing_key', value=None)

        with self.assertRaises(RheaError):
            rhea_parser.get_dict(key='dict_non_existing_key', value=NO_VALUE_FOUND)

        with self.assertRaises(RheaError):
            rhea_parser.get_dict(key='dict_non_existing_key', value=None, is_list=True)

        self.assertEqual(
            rhea_parser.get_dict(key='dict_non_existing_key', value=None, is_optional=True), None)
        self.assertEqual(rhea_parser.get_dict(key='dict_non_existing_key',
                                              value=None,
                                              is_optional=True,
                                              default={'foo': 'bar'}), {'foo': 'bar'})

        self.assertEqual(rhea_parser.get_dict(key='dict_non_existing_key',
                                              value=None,
                                              is_list=True,
                                              is_optional=True), None)
        self.assertEqual(rhea_parser.get_dict(key='dict_non_existing_key',
                                              value=None,
                                              is_list=True,
                                              is_optional=True,
                                              default=[{'foo': 'bar'}, {'foo': 'boo'}]),
                         [{'foo': 'bar'}, {'foo': 'boo'}])

    def test_get_uri(self):
        value = rhea_parser.get_uri(key='uri_key_1', value="user:pass@siteweb.ca")
        self.assertEqual(value, UriSpec("user", "pass", "siteweb.ca"))

        value = rhea_parser.get_uri(key='uri_key_2', value="user2:pass@localhost:8080")
        self.assertEqual(value, UriSpec("user2", "pass", "localhost:8080"))

        value = rhea_parser.get_uri(key='uri_key_3', value="user2:pass@https://quay.io")
        self.assertEqual(value, UriSpec("user2", "pass", "https://quay.io"))

        value = rhea_parser.get_uri(key='uri_list_key_1',
                                    value=[
                                        "user:pass@siteweb.ca",
                                        "user2:pass@localhost:8080",
                                        "user2:pass@https://quay.io"],
                                    is_list=True)
        self.assertEqual(value, [
            UriSpec("user", "pass", "siteweb.ca"),
            UriSpec("user2", "pass", "localhost:8080"),
            UriSpec("user2", "pass", "https://quay.io")
        ])

        with self.assertRaises(RheaError):
            rhea_parser.get_uri(key='uri_error_key_1', value="foo")

        with self.assertRaises(RheaError):
            rhea_parser.get_uri(key='uri_error_key_2', value=1)

        with self.assertRaises(RheaError):
            rhea_parser.get_uri(key='uri_error_key_3', value=False)

        with self.assertRaises(RheaError):
            rhea_parser.get_uri(key='uri_error_key_4', value=["1", "foo"])

        with self.assertRaises(RheaError):
            rhea_parser.get_uri(key='uri_list_key_1',
                                value=[
                                    "user:pass@siteweb.ca",
                                    "user2:pass@localhost:8080",
                                    "user2:pass@https://quay.io"])

        with self.assertRaises(RheaError):
            rhea_parser.get_uri(key='uri_list_error_key_1',
                                value=["123", "user:pass@siteweb.ca"],
                                is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_uri(key='uri_list_error_key_2',
                                value=["user:pass@siteweb.ca", 12.3],
                                is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_uri(key='uri_list_error_key_3',
                                value=["user:pass@siteweb.ca", None],
                                is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_uri(key='uri_list_error_key_4',
                                value=["user:pass@siteweb.ca", "123", False],
                                is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_uri(key='uri_key_1', value="user:pass@siteweb.ca", is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_uri(key='uri_non_existing_key', value=None)

        with self.assertRaises(RheaError):
            rhea_parser.get_uri(key='uri_non_existing_key', value=NO_VALUE_FOUND)

        with self.assertRaises(RheaError):
            rhea_parser.get_uri(key='uri_non_existing_key', value=None, is_list=True)

        self.assertEqual(rhea_parser.get_uri(key='uri_non_existing_key',
                                             value=None,
                                             is_optional=True), None)
        self.assertEqual(
            rhea_parser.get_uri(key='uri_non_existing_key',
                                value=None,
                                is_optional=True,
                                default=UriSpec("user2", "pass", "localhost:8080")),
            UriSpec("user2", "pass", "localhost:8080"))

        self.assertEqual(
            rhea_parser.get_uri(key='uri_non_existing_key',
                                value=None,
                                is_list=True,
                                is_optional=True), None)
        self.assertEqual(
            rhea_parser.get_uri(key='uri_non_existing_key',
                                value=None,
                                is_list=True,
                                is_optional=True,
                                default=[UriSpec("user", "pass", "siteweb.ca"),
                                         UriSpec("user2", "pass", "localhost:8080")]),
            [UriSpec("user", "pass", "siteweb.ca"), UriSpec("user2", "pass", "localhost:8080")])

    def test_get_auth(self):
        value = rhea_parser.get_auth(key='auth_key_1', value="user:pass")
        self.assertEqual(value, AuthSpec("user", "pass"))

        value = rhea_parser.get_auth(key='auth_list_key_1',
                                     value=[
                                         "user:pass",
                                         "user2:pass"],
                                     is_list=True)
        self.assertEqual(value, [
            AuthSpec("user", "pass"),
            AuthSpec("user2", "pass"),
        ])

        with self.assertRaises(RheaError):
            rhea_parser.get_auth(key='auth_error_key_1', value="foo")

        with self.assertRaises(RheaError):
            rhea_parser.get_auth(key='auth_error_key_2', value=1)

        with self.assertRaises(RheaError):
            rhea_parser.get_auth(key='auth_error_key_3', value=False)

        with self.assertRaises(RheaError):
            rhea_parser.get_auth(key='auth_error_key_4', value=["1", "foo"])

        with self.assertRaises(RheaError):
            rhea_parser.get_auth(key='auth_list_key_1',
                                 value=[
                                     "user:pass",
                                     "user2:pass"])

        with self.assertRaises(RheaError):
            rhea_parser.get_auth(key='auth_list_error_key_1',
                                 value=["123", "user:pass"],
                                 is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_auth(key='auth_list_error_key_2',
                                 value=["user:pass", 12.3],
                                 is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_auth(key='auth_list_error_key_3',
                                 value=["user:pass", None],
                                 is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_auth(key='auth_list_error_key_4',
                                 value=["user:pass", "123", False],
                                 is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_auth(key='auth_key_1',
                                 value="user:pass",
                                 is_list=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_auth(key='auth_non_existing_key', value=None)

        with self.assertRaises(RheaError):
            rhea_parser.get_auth(key='auth_non_existing_key', value=NO_VALUE_FOUND)

        with self.assertRaises(RheaError):
            rhea_parser.get_auth(key='auth_non_existing_key', value=None, is_list=True)

        self.assertEqual(
            rhea_parser.get_auth(key='auth_non_existing_key', value=None, is_optional=True), None)
        self.assertEqual(
            rhea_parser.get_auth(key='auth_non_existing_key',
                                 value=None,
                                 is_optional=True,
                                 default=AuthSpec("user2", "pass")),
            AuthSpec("user2", "pass"))

        self.assertEqual(
            rhea_parser.get_auth(key='auth_non_existing_key',
                                 value=None,
                                 is_list=True,
                                 is_optional=True), None)
        self.assertEqual(
            rhea_parser.get_auth(key='auth_non_existing_key',
                                 value=None,
                                 is_list=True,
                                 is_optional=True,
                                 default=[AuthSpec("user", "pass"),
                                          AuthSpec("user2", "pass")]),
            [AuthSpec("user", "pass"), AuthSpec("user2", "pass")])

    def test_get_list(self):
        value = rhea_parser.get_list(key='list_key_1',
                                     value="user:pass@siteweb.ca, 'pp', 0.1, 'foo'")
        self.assertEqual(value, ['user:pass@siteweb.ca', "'pp'", '0.1', "'foo'"])

        value = rhea_parser.get_list(key='list_key_2',
                                     value="user1,user2 , user3,     user4    , user5")
        self.assertEqual(value, ['user1', 'user2', 'user3', 'user4', 'user5'])

        value = rhea_parser.get_list(key='list_key_3', value=[False])
        self.assertEqual(value, [False])

        value = rhea_parser.get_list(key='list_key_3', value=['false'])
        self.assertEqual(value, ['false'])

        value = rhea_parser.get_list(key='list_key_4', value="foo")
        self.assertEqual(value, ['foo'])

        value = rhea_parser.get_list(key='list_key_5', value="")
        self.assertEqual(value, [])

        value = rhea_parser.get_list(key='list_error_key_3', value='null')
        self.assertEqual(value, ['null'])

        with self.assertRaises(RheaError):
            rhea_parser.get_list(key='list_error_key_1', value=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_list(key='list_error_key_2', value={"key": "value"})

        with self.assertRaises(RheaError):
            rhea_parser.get_list(key='list_error_key_4', value=123)

        with self.assertRaises(RheaError):
            rhea_parser.get_list(key='list_non_existing_key', value=None)

        with self.assertRaises(RheaError):
            rhea_parser.get_list(key='list_non_existing_key', value=NO_VALUE_FOUND)

        self.assertEqual(rhea_parser.get_list(key='list_non_existing_key',
                                              value=None,
                                              is_optional=True), None)
        self.assertEqual(rhea_parser.get_list(key='list_non_existing_key',
                                              value=None,
                                              is_optional=True,
                                              default=['foo']), ['foo'])

    def test_get_dict_of_dicts(self):
        value = rhea_parser.get_dict_of_dicts(
            key='dict_dicts_key_1',
            value={"data1": {"mountPath": "/data/21", "existingClaim": "data-1-pvc"}})
        self.assertEqual(value, {'data1': {'mountPath': '/data/21', 'existingClaim': 'data-1-pvc'}})

        value = rhea_parser.get_dict_of_dicts(
            key='dict_dicts_key_1',
            value='{"data1": {"mountPath": "/data/21", "existingClaim": "data-1-pvc"}}')
        self.assertEqual(value, {'data1': {'mountPath': '/data/21', 'existingClaim': 'data-1-pvc'}})

        value = rhea_parser.get_dict_of_dicts(
            key='dict_dicts_key_2',
            value={
                "outputs1": {"mountPath": "/output/2", "existingClaim": "outputs-1-pvc"},
                "outputs2": {"mountPath": "/output/2", "existingClaim": "output-2-pvc"}})
        self.assertEqual(value, {
            'outputs1': {'mountPath': '/output/2', 'existingClaim': 'outputs-1-pvc'},
            'outputs2': {'mountPath': '/output/2', 'existingClaim': 'output-2-pvc'}})

        value = rhea_parser.get_dict_of_dicts(key='dict_dicts_key_3', value={})
        self.assertEqual(value, None)

        with self.assertRaises(RheaError):
            rhea_parser.get_dict_of_dicts(key='dict_dicts_error_key_1', value=True)

        with self.assertRaises(RheaError):
            rhea_parser.get_dict_of_dicts(key='dict_dicts_error_key_2', value={"key": "value"})

        with self.assertRaises(RheaError):
            rhea_parser.get_dict_of_dicts(key='dict_dicts_error_key_3', value='null')

        with self.assertRaises(RheaError):
            rhea_parser.get_dict_of_dicts(key='dict_dicts_error_key_4', value=123)

        with self.assertRaises(RheaError):
            rhea_parser.get_dict_of_dicts(key='dict_dicts_error_key_5', value="foo")

        with self.assertRaises(RheaError):
            rhea_parser.get_dict_of_dicts(key='dict_dicts_error_key_6', value="")

        with self.assertRaises(RheaError):
            rhea_parser.get_dict_of_dicts(
                key='dict_dicts_error_key_7',
                value={"mountPath": "/data/2", "existingClaim": "data-2-pvc"})

        with self.assertRaises(RheaError):
            rhea_parser.get_dict_of_dicts(key='dict_dicts_non_existing_key', value=None)

        with self.assertRaises(RheaError):
            rhea_parser.get_dict_of_dicts(key='dict_dicts_non_existing_key', value=NO_VALUE_FOUND)

        self.assertEqual(rhea_parser.get_dict_of_dicts(key='dict_dicts_non_existing_key',
                                                       value=None,
                                                       is_optional=True), None)
        self.assertEqual(rhea_parser.get_dict_of_dicts(key='dict_dicts_non_existing_key',
                                                       value=None,
                                                       is_optional=True, default={}), {})

    def test_get_wasbs_path(self):
        # Correct url
        wasbs_path = 'wasbs://container@user.blob.core.windows.net/path'
        parsed_url = rhea_parser.parse_wasbs_path(wasbs_path)
        assert parsed_url == WasbsSpec('container', 'user', 'path')
        wasbs_path = 'wasbs://container@user.blob.core.windows.net/'
        parsed_url = rhea_parser.parse_wasbs_path(wasbs_path)
        assert parsed_url == WasbsSpec('container', 'user', '')
        wasbs_path = 'wasbs://container@user.blob.core.windows.net'
        parsed_url = rhea_parser.parse_wasbs_path(wasbs_path)
        assert parsed_url == WasbsSpec('container', 'user', '')
        wasbs_path = 'wasbs://container@user.blob.core.windows.net/path/to/file'
        parsed_url = rhea_parser.parse_wasbs_path(wasbs_path)
        assert parsed_url == WasbsSpec('container', 'user', 'path/to/file')

        # Wrong url
        wasbs_path = 'wasbs://container@user.foo.bar.windows.net/path/to/file'
        with self.assertRaises(RheaError):
            rhea_parser.parse_wasbs_path(wasbs_path)

        wasbs_path = 'wasbs://container@user.blob.core.foo.net/path/to/file'
        with self.assertRaises(RheaError):
            rhea_parser.parse_wasbs_path(wasbs_path)

        wasbs_path = 'wasbs://container@user.blob.windows.net/path/to/file'
        with self.assertRaises(RheaError):
            rhea_parser.parse_wasbs_path(wasbs_path)

    def test_parse_gcs_path(self):
        # Correct url
        gcs_path = 'gs://bucket/path/to/blob'
        parsed_url = rhea_parser.parse_gcs_path(gcs_path)
        assert parsed_url == GCSSpec('bucket', 'path/to/blob')

        # Wrong url
        gcs_path = 'gs:/bucket/path/to/blob'
        with self.assertRaises(RheaError):
            rhea_parser.parse_gcs_path(gcs_path)

        # Trailing slash
        gcs_path = 'gs://bucket/path/to/blob/'
        assert rhea_parser.parse_gcs_path(gcs_path) == GCSSpec('bucket', 'path/to/blob/')

        # Bucket only
        gcs_path = 'gs://bucket/'
        assert rhea_parser.parse_gcs_path(gcs_path) == GCSSpec('bucket', '')

    def test_parse_s3_path(self):
        s3_path = 's3://test/this/is/bad/key.txt'
        parsed_url = rhea_parser.parse_s3_path(s3_path)
        assert parsed_url == S3Spec('test', 'this/is/bad/key.txt')



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from rhea import parser, reader
from rhea.constants import NO_VALUE_FOUND


class Rhea(object):
    def __init__(self, **data):
        self._data = data
        self._requested_keys = set()
        self._secret_keys = set()
        self._local_keys = set()

    @classmethod
    def read_configs(cls, config_values):  # pylint:disable=redefined-outer-name
        config = reader.read(config_values)  # pylint:disable=redefined-outer-name
        return cls(**config) if config else None

    def keys_startswith(self, term):
        return [k for k in self._data if k.startswith(term)]

    def keys_endswith(self, term):
        return [k for k in self._data if k.endswith(term)]

    def has_key(self, key):
        return key in self._data

    @property
    def data(self):
        return self._data

    @property
    def requested_keys(self):
        return self._requested_keys

    @property
    def secret_keys(self):
        return self._secret_keys

    @property
    def local_keys(self):
        return self._local_keys

    def get_requested_data(self, include_secrets=False, include_locals=False, to_str=False):
        data = {}
        for key in self._requested_keys:
            if not include_secrets and key in self._secret_keys:
                continue
            if not include_locals and key in self._local_keys:
                continue
            value = self._data[key]
            data[key] = '{}'.format(value) if to_str else value
        return data

    def get_int(self,
                key,
                is_list=False,
                is_optional=False,
                is_secret=False,
                is_local=False,
                default=None,
                options=None):
        """
        Get a the value corresponding to the key and converts it to `int`/`list(int)`.

        Args:
            key: the dict key.
            is_list: If this is one element or a list of elements.
            is_optional: To raise an error if key was not found.
            is_secret: If the key is a secret.
            is_local: If the key is a local to this service.
            default: default value if is_optional is True.
            options: list/tuple if provided, the value must be one of these values.

        Returns:
             `int`: value corresponding to the key.
        """
        return self._get(key=key,
                         parser_fct=parser.get_int,
                         is_list=is_list,
                         is_optional=is_optional,
                         is_secret=is_secret,
                         is_local=is_local,
                         default=default,
                         options=options)

    def get_float(self,
                  key,
                  is_list=False,
                  is_optional=False,
                  is_secret=False,
                  is_local=False,
                  default=None,
                  options=None):
        """
        Get a the value corresponding to the key and converts it to `float`/`list(float)`.

        Args:
            key: the dict key.
            is_list: If this is one element or a list of elements.
            is_optional: To raise an error if key was not found.
            is_secret: If the key is a secret.
            is_local: If the key is a local to this service.
            default: default value if is_optional is True.
            options: list/tuple if provided, the value must be one of these values.

        Returns:
             `float`: value corresponding to the key.
        """
        return self._get(key=key,
                         parser_fct=parser.get_float,
                         is_list=is_list,
                         is_optional=is_optional,
                         is_secret=is_secret,
                         is_local=is_local,
                         default=default,
                         options=options)

    def get_boolean(self,
                    key,
                    is_list=False,
                    is_optional=False,
                    is_secret=False,
                    is_local=False,
                    default=None,
                    options=None):
        """
        Get a the value corresponding to the key and converts it to `bool`/`list(str)`.

        Args:
            key: the dict key.
            is_list: If this is one element or a list of elements.
            is_optional: To raise an error if key was not found.
            is_secret: If the key is a secret.
            is_local: If the key is a local to this service.
            default: default value if is_optional is True.
            options: list/tuple if provided, the value must be one of these values.

        Returns:
            `bool`: value corresponding to the key.
        """
        return self._get(key=key,
                         parser_fct=parser.get_boolean,
                         is_list=is_list,
                         is_optional=is_optional,
                         is_secret=is_secret,
                         is_local=is_local,
                         default=default,
                         options=options)

    def get_string(self,
                   key,
                   is_list=False,
                   is_optional=False,
                   is_secret=False,
                   is_local=False,
                   default=None,
                   options=None):
        """
        Get a the value corresponding to the key and converts it to `str`/`list(str)`.

        Args:
            key: the dict key.
            is_list: If this is one element or a list of elements.
            is_optional: To raise an error if key was not found.
            is_secret: If the key is a secret.
            is_local: If the key is a local to this service.
            default: default value if is_optional is True.
            options: list/tuple if provided, the value must be one of these values.

        Returns:
            `str`: value corresponding to the key.
        """
        return self._get(key=key,
                         parser_fct=parser.get_string,
                         is_list=is_list,
                         is_optional=is_optional,
                         is_secret=is_secret,
                         is_local=is_local,
                         default=default,
                         options=options)

    def get_dict(self,
                 key,
                 is_list=False,
                 is_optional=False,
                 is_secret=False,
                 is_local=False,
                 default=None,
                 options=None):
        """
        Get a the value corresponding to the key and converts it to `dict`.

        Args:
            key: the dict key.
            is_list: If this is one element or a list of elements.
            is_optional: To raise an error if key was not found.
            is_secret: If the key is a secret.
            is_local: If the key is a local to this service.
            default: default value if is_optional is True.
            options: list/tuple if provided, the value must be one of these values.

        Returns:
            `str`: value corresponding to the key.
        """

        return self._get(key=key,
                         parser_fct=parser.get_dict,
                         is_list=is_list,
                         is_optional=is_optional,
                         is_secret=is_secret,
                         is_local=is_local,
                         default=default,
                         options=options)

    def get_dict_of_dicts(self,
                          key,
                          is_optional=False,
                          is_secret=False,
                          is_local=False,
                          default=None,
                          options=None):
        """
        Get a the value corresponding to the key and converts it to `dict`.

        Add an extra validation that all keys have a dict as values.

        Args:
            key: the dict key.
            is_optional: To raise an error if key was not found.
            is_secret: If the key is a secret.
            is_local: If the key is a local to this service.
            default: default value if is_optional is True.
            options: list/tuple if provided, the value must be one of these values.

        Returns:
            `str`: value corresponding to the key.
        """
        return self._get(key=key,
                         parser_fct=parser.get_dict_of_dicts,
                         is_list=None,
                         is_optional=is_optional,
                         is_secret=is_secret,
                         is_local=is_local,
                         default=default,
                         options=options)

    def get_uri(self,
                key,
                is_list=False,
                is_optional=False,
                is_secret=False,
                is_local=False,
                default=None,
                options=None):
        """
        Get a the value corresponding to the key and converts it to `UriSpec`.

        Args
            key: the dict key.
            is_list: If this is one element or a list of elements.
            is_optional: To raise an error if key was not found.
            is_secret: If the key is a secret.
            is_local: If the key is a local to this service.
            default: default value if is_optional is True.
            options: list/tuple if provided, the value must be one of these values.

        Returns:
             `str`: value corresponding to the key.
        """
        return self._get(key=key,
                         parser_fct=parser.get_uri,
                         is_list=is_list,
                         is_optional=is_optional,
                         is_secret=is_secret,
                         is_local=is_local,
                         default=default,
                         options=options)

    def get_auth(self,
                 key,
                 is_list=False,
                 is_optional=False,
                 is_secret=False,
                 is_local=False,
                 default=None,
                 options=None):
        """
        Get a the value corresponding to the key and converts it to `AuthSpec`.

        Args
            key: the dict key.
            is_list: If this is one element or a list of elements.
            is_optional: To raise an error if key was not found.
            is_secret: If the key is a secret.
            is_local: If the key is a local to this service.
            default: default value if is_optional is True.
            options: list/tuple if provided, the value must be one of these values.

        Returns:
             `str`: value corresponding to the key.
        """
        return self._get(key=key,
                         parser_fct=parser.get_auth,
                         is_list=is_list,
                         is_optional=is_optional,
                         is_secret=is_secret,
                         is_local=is_local,
                         default=default,
                         options=options)

    def get_list(self,
                 key,
                 is_optional=False,
                 is_secret=False,
                 is_local=False,
                 default=None,
                 options=None):
        """
        Get a the value corresponding to the key and converts comma separated values to a list.

        Args:
            key: the dict key.
            is_optional: To raise an error if key was not found.
            is_secret: If the key is a secret.
            is_local: If the key is a local to this service.
            default: default value if is_optional is True.
            options: list/tuple if provided, the value must be one of these values.

        Returns:
             `str`: value corresponding to the key.
        """

        return self._get(key=key,
                         parser_fct=parser.get_list,
                         is_list=None,
                         is_optional=is_optional,
                         is_secret=is_secret,
                         is_local=is_local,
                         default=default,
                         options=options)

    def _get(self, key, parser_fct, is_list, is_optional, is_secret, is_local, default, options):
        """
        Get key from the dictionary made out of the configs passed.

        Args:
            key: the dict key.

        Returns:
             The corresponding value of the key if found.

        Raises:
            KeyError
        """
        value = self._data.get(key, NO_VALUE_FOUND)
        parsed_value = parser_fct(key=key,
                                  value=value,
                                  is_list=is_list,
                                  is_optional=is_optional,
                                  default=default,
                                  options=options)
        self._add_key(key, is_secret=is_secret, is_local=is_local)
        return parsed_value

    def _add_key(self, key, is_secret=False, is_local=False):
        self._requested_keys.add(key)
        if is_secret:
            self._secret_keys.add(key)
        if is_local:
            self._local_keys.add(key)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from collections import namedtuple


class UriSpec(namedtuple("UriSpec", "user password host")):
    """
    A specification for uris configuration.
    """
    pass


class AuthSpec(namedtuple("UriSpec", "user password")):
    """
    A specification for auth configuration.
    """
    pass


class WasbsSpec(namedtuple("WasbsSpec", "container storage_account path")):
    """
    A specification for wasbs configuration.
    """


class GCSSpec(namedtuple("GCSSpec", "bucket blob")):
    """
    A specification for gcs configuration.
    """


class S3Spec(namedtuple("S3Spec", "bucket key")):
    """
    A specification for s3 configuration.
    """



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

from collections import Mapping

from rhea.config import ConfigSpec
from rhea.exceptions import RheaError
from rhea.utils import deep_update, to_list


def read(config_values):
    """Reads an ordered list of configuration values and deep merge the values in reverse order."""
    if not config_values:
        raise RheaError('Cannot read config_value: `{}`'.format(config_values))

    config_values = to_list(config_values)

    config = {}
    for config_value in config_values:
        config_value = ConfigSpec.get_from(value=config_value)
        config_value.check_type()
        config_results = config_value.read()
        if config_results and isinstance(config_results, Mapping):
            config = deep_update(config, config_results)
        elif config_value.check_if_exists:
            raise RheaError('Cannot read config_value: `{}`'.format(config_value))

    return config
